from urllib.parse import quote

from odm_sdk import BioMetaKeys, GenestackException, Metainfo
from odm_sdk.metainfo_scalar_values import ExternalLink, FileReference, StringValue

ANNOTATION_KEY = 'genestack.url:annotations'
SEQUENCE_KEY = 'genestack.url:sequence'


class DataImporter(object):
    """
    A class used to import files to a Genestack instance.
    If no ``parent`` is specified, the files are created in the special folder ``Imported files``

    Required and recommended values can be set by arguments directly
    or passed inside a :py:class:`~odm_sdk.Metainfo`::

       create_bed(name="Bed", url="some/url")

       # is equivalent to:
       metainfo = Metainfo()
       metainfo.add_string(Metainfo.NAME, "Bed")
       metainfo.add_external_link(Metainfo.DATA_LINK, "some/url", text="link name")
       create_bed(metainfo=metainfo)

    However, **do not** pass the same value both through the arguments and inside a metainfo object.

    Genestack accepts both compressed and uncompressed files.
    ``file://`` is used for files that are local to the ODM instance.
    Special characters should be escaped except ``s3://``. Links to Amazon S3 storage should be formatted
    as in `s3cmd <http://s3tools.org/s3cmd>`__.

    Supported protocols:

    * ``file://``:
        - ``file://test.txt``

    * ``ftp://``
        - ``ftp://server.com/file.txt``

    * ``http://`` ``https://``
        - ``http://server.com/file.txt``

    * ``ascp://``
        - ``ascp://<user>@<server>:file.txt``

    * ``s3://``
        -  ``s3://bucket/file.gz``
        -  ``s3://bucket/file name.gz``

    """

    #: Affymetrix microarray annotation type
    AFFYMETRIX_ANNOTATION = 'affymetrixMicroarrayAnnotation'
    #: Agilent microarray annotation type
    AGILENT_ANNOTATION = 'agilentMicroarrayAnnotation'
    #: TSV (GenePix etc) microarray annotation type
    TSV_ANNOTATION = 'TSVMicroarrayAnnotation'
    #: Infinium microarray annotation type
    INFINIUM_ANNOTATION = 'methylationArrayAnnotation'

    #: Supported microarray annotation types
    MICROARRAY_ANNOTATION_TYPES = (
        AGILENT_ANNOTATION,
        AFFYMETRIX_ANNOTATION,
        TSV_ANNOTATION,
        INFINIUM_ANNOTATION
    )

    def __init__(self, connection):
        self.connection = connection
        self.importer = connection.application('genestack/upload')

    def __process_links(self, metainfo):
        all_links = [(key, val) for key, val in metainfo.items() if val[0]['type'] == 'externalLink']
        for key, external_link_list in all_links:
            for x in external_link_list:
                if x['url'].startswith('s3://'):
                    x['url'] = 's3://%s' % quote(x['url'][5:])

    def __invoke_loader(self, parent, importer_type, metainfo):
        self.__process_links(metainfo)
        return self.importer.invoke('createFile', parent, importer_type, metainfo)['accession']

    @staticmethod
    def __add_to_metainfo(metainfo, key, value, value_type, required=False):
        """
        Add ``key``: ``value`` pair to ``metainfo``.

        Fails if:
        * ``metainfo`` already contains ``key``;
        * ``key`` is ``required``, but no value is provided by user,
          and neither ``metainfo`` has already existing ``key`` (with some value).

        :param metainfo: metainfo object to be updated
        :type metainfo: Metainfo
        :param key: key to add
        :type key: str
        :param value: single argument to pass to the value class constructor, or list if need to add many values.
        :type value: object
        :param value_type: value class
        :type value_type: T => odm_sdk.metainfo_scalar_values.MetainfoScalarValue
        :param required: flag if value is required
        :type required: bool
        :return: None
        """
        current_value = metainfo.get(key)
        if current_value is not None and value is not None:
            raise GenestackException('Key "%s" is passed both as function argument '
                                     'and inside metainfo object' % key)
        if current_value:
            return

        if required and value is None:
            raise GenestackException('Missing required key "%s", '
                                     'it should be passed as function argument '
                                     'or in metainfo object' % key)
        if value is not None:
            value_list = value if isinstance(value, list) else [value]
            for val in value_list:
                metainfo.add_value(key, value_type(val))

    def create_bed(self, parent=None, name=None, reference_genome=None, url=None, metainfo=None):
        """
        Create a Genestack BED Track from a local or remote BED file.
        ``name`` and ``url`` are mandatory fields. They can be specified through the arguments or
        via a :py:class:`~odm_sdk.Metainfo` instance.

        :param parent: accession of parent folder
            (if not provided, files will be created in the ``Imported files`` folder)
        :type parent: str
        :param name: name of the file
        :type name: str
        :param reference_genome: accession of reference genome
        :type reference_genome: str
        :param url: URL or local path to file
        :type url: str
        :param metainfo: metainfo object
        :type metainfo: Metainfo
        :return: file accession
        :rtype: str
        """
        metainfo = DataImporter._copy_metainfo(metainfo)
        self.__add_to_metainfo(metainfo, Metainfo.NAME, name, StringValue, required=True)
        self.__add_to_metainfo(metainfo, BioMetaKeys.REFERENCE_GENOME,
                               reference_genome, FileReference)
        self.__add_to_metainfo(metainfo, BioMetaKeys.DATA_LINK, url, ExternalLink, required=True)
        return self.__invoke_loader(parent, 'bedFiles', metainfo)

    def create_vcf(self, parent=None, name=None, reference_genome=None, url=None, metainfo=None):
        """
        Create a Genestack Variants file from a local or remote VCF file.
        ``name`` and ``url`` are required fields. They can be specified through the arguments or
        via a :py:class:`~odm_sdk.Metainfo` instance.

        :param parent: accession of parent folder
            (if not provided, files will be created in the ``Imported files`` folder)
        :type parent: str
        :param name: name of the file
        :type name: str
        :param reference_genome: accession of reference genome
        :type reference_genome: str
        :param url: URL or local path to file
        :type url: str
        :param metainfo: metainfo object
        :type metainfo: Metainfo
        :return: file accession
        :rtype: str
        """
        metainfo = DataImporter._copy_metainfo(metainfo)
        self.__add_to_metainfo(metainfo, Metainfo.NAME, name, StringValue, required=True)
        self.__add_to_metainfo(metainfo, BioMetaKeys.REFERENCE_GENOME,
                               reference_genome, FileReference)
        self.__add_to_metainfo(metainfo, BioMetaKeys.DATA_LINK, url, ExternalLink, required=True)
        return self.__invoke_loader(parent, 'variationFiles', metainfo)

    def create_wig(self, parent=None, name=None, reference_genome=None, url=None, metainfo=None):
        """
        Create a Genestack Wiggle Track from a local or remote WIG file.
        ``name`` and ``url`` are required fields.
        They can be specified through the arguments or
        via a :py:class:`~odm_sdk.Metainfo` instance.

        :param parent: accession of parent folder
            (if not provided, files will be created in the ``Imported files`` folder)
        :type parent: str
        :param name: name of the file
        :type name: str
        :param reference_genome: accession of reference genome
        :type reference_genome: str
        :param url: URL or local path to file
        :type url: str
        :param metainfo: metainfo object
        :type metainfo: Metainfo
        :return: file accession
        :rtype: str
        """
        metainfo = DataImporter._copy_metainfo(metainfo)
        self.__add_to_metainfo(metainfo, Metainfo.NAME, name, StringValue, required=True)
        self.__add_to_metainfo(metainfo, BioMetaKeys.REFERENCE_GENOME,
                               reference_genome, FileReference)
        self.__add_to_metainfo(metainfo, BioMetaKeys.DATA_LINK, url, ExternalLink, required=True)
        return self.__invoke_loader(parent, 'wigFiles', metainfo)

    def create_bam(self,
                   parent=None,
                   name=None,
                   url=None,
                   organism=None,
                   strain=None,
                   reference_genome=None,
                   metainfo=None,
                   ):
        """
        Create a Genestack Aligned Reads file from a local or remote BAM file.
        ``name``, ``url`` and ``organism`` are required fields.
        They can be specified through the arguments or
        via a :py:class:`~odm_sdk.Metainfo` instance.


        :param parent: accession of parent folder
            (if not provided, files will be created in the ``Imported files`` folder)
        :type parent: str
        :param name: name of the file
        :type name: str
        :param url: URL of a BAM file; the index will be created at initialization
        :param organism: organism
        :type organism: str
        :param strain: strain
        :type strain:
        :param reference_genome: reference genome accession
        :type reference_genome: str
        :type metainfo: Metainfo
        :param metainfo: metainfo object
        :return: file accession
        :rtype: str
        """
        metainfo = DataImporter._copy_metainfo(metainfo)
        self.__add_to_metainfo(metainfo, Metainfo.NAME, name, StringValue, required=True)
        self.__add_to_metainfo(metainfo, BioMetaKeys.ORGANISM, organism, StringValue, required=True)
        self.__add_to_metainfo(metainfo, BioMetaKeys.STRAIN, strain, StringValue)
        self.__add_to_metainfo(metainfo, BioMetaKeys.REFERENCE_GENOME, reference_genome, FileReference)
        self.__add_to_metainfo(metainfo, BioMetaKeys.BAM_FILE_LINK, url, ExternalLink, required=True)
        return self.__invoke_loader(parent, 'alignedReads', metainfo)

    @staticmethod
    def _copy_metainfo(metainfo):
        return Metainfo(metainfo) if metainfo else Metainfo()

    def create_experiment(self, parent=None, name=None, description=None, metainfo=None):
        raise GenestackException('"create_experiment" is not available anymore, '
                                 'use "FilesUtil.create_dataset" method')

    def create_microarray_assay(self, parent, name=None, urls=None,
                                method=None, organism=None, metainfo=None):
        raise GenestackException('"create_microarray_assay" is not available anymore, '
                                 'use "create_microarray_data" method')

    def create_microarray_data(self, parent, name=None, urls=None,
                               method=None, organism=None, metainfo=None):
        """
        Create a Genestack Microarray Data inside an folder.
        ``name`` and ``urls`` are required fields.
        They can be specified through the arguments or
        via a :py:class:`~odm_sdk.Metainfo` instance.

        :param parent: accession of parent folder
        :type parent: str
        :param name: name of the file
        :type name: str
        :param urls: list of urls
        :type urls: list
        :param method: method
        :type method: str
        :param organism: organism
        :type organism: str
        :param metainfo: metainfo object
        :type metainfo: Metainfo
        :return: file accession
        :rtype: str
        """
        metainfo = DataImporter._copy_metainfo(metainfo)
        self.__add_to_metainfo(metainfo, Metainfo.NAME, name, StringValue, required=True)
        self.__add_to_metainfo(metainfo, BioMetaKeys.ORGANISM, organism, StringValue)
        self.__add_to_metainfo(metainfo, BioMetaKeys.METHOD, method, StringValue)
        self.__add_to_metainfo(metainfo, BioMetaKeys.DATA_LINK, urls, ExternalLink, required=True)
        return self.__invoke_loader(parent, 'microarrayData', metainfo)

    def create_infinium_microarray_data(self, parent, name=None, urls=None,
                                        method=None, metainfo=None):
        """
        Create a Genestack Infinium Microarrays Data inside a folder. We can't use create_microarray_data method
        because 'microarrayData' importer can have only one source file, while infinium assay has two. So we invoke
        'infinium MicroarrayData' importer with two links for BioMetaKeys.DATA_LINK key in metainfo.

        Infinum microarrays available only for humans so we have no 'organism' key in arguments.

        :param parent: accession of parent folder
        :type parent: str
        :param name: name of the file
        :type name: str
        :param urls: list of urls
        :type urls: list
        :param method: method
        :type method: str
        :param metainfo: metainfo object
        :type metainfo: Metainfo
        :return: file accession
        :rtype: str
        """
        metainfo = DataImporter._copy_metainfo(metainfo)
        self.__add_to_metainfo(metainfo, Metainfo.NAME, name, StringValue, required=True)
        self.__add_to_metainfo(metainfo, BioMetaKeys.METHOD, method, StringValue)
        self.__add_to_metainfo(metainfo, BioMetaKeys.DATA_LINK, urls, ExternalLink, required=True)
        return self.__invoke_loader(parent, 'infiniumMicroarrayData', metainfo)

    def create_sequencing_assay(self, parent, name=None, urls=None,
                                method=None, organism=None, metainfo=None):
        raise GenestackException('"create_sequencing_assay" is not available anymore, '
                                 'use "create_unaligned_read" method')

    def create_unaligned_read(self, parent=None, name=None, urls=None,
                              method=None, organism=None, metainfo=None):
        """
        Create a Genestack Unaligned Reads file from one or several local or remote files.
        Most common file formats encoding
        sequencing reads with quality scores are accepted (FASTQ 33/64, SRA, FASTA+QUAL, SFF, FAST5).
        ``name`` and ``urls`` are required fields.
        They can be specified through the arguments or via a :py:class:`~odm_sdk.Metainfo` instance.

        :param parent: accession of parent folder
            (if not provided, files will be created in the ``Imported files`` folder)
        :type parent: str
        :param name: name of the file
        :type name: str
        :param urls: list of urls
        :type urls: list
        :param method: method
        :type method: str
        :param organism: organism
        :type organism: str
        :param metainfo: metainfo object
        :type metainfo: Metainfo
        :return: file accession
        :rtype: str
        """
        metainfo = DataImporter._copy_metainfo(metainfo)
        self.__add_to_metainfo(metainfo, Metainfo.NAME, name, StringValue, required=True)
        self.__add_to_metainfo(metainfo, BioMetaKeys.ORGANISM, organism, StringValue)
        self.__add_to_metainfo(metainfo, BioMetaKeys.METHOD, method, StringValue)
        self.__add_to_metainfo(metainfo, BioMetaKeys.READS_LINK, urls, ExternalLink, required=True)
        return self.__invoke_loader(parent, 'rawReads', metainfo)

    def create_genome_annotation(self, parent=None, url=None, name=None,
                                 organism=None, reference_genome=None,
                                 strain=None, metainfo=None):
        """
        Create a Genestack Genome Annotation file from a local or remote file.
        ``name`` and ``url`` are required fields.
        They can be specified through the arguments or
        via a :py:class:`~odm_sdk.Metainfo` instance.

        :param parent: accession of parent folder
            (if not provided, files will be created in the ``Imported files`` folder)
        :type parent: str
        :param url: URL or local path
        :type url: str
        :param name: name of the file
        :type name: str
        :param organism: organism
        :type organism: str
        :param reference_genome: reference genome accession
        :type reference_genome: str
        :param strain: strain
        :type strain: str
        :param metainfo: metainfo object
        :type metainfo: Metainfo
        :return: file accession
        :rtype: str
        """
        metainfo = DataImporter._copy_metainfo(metainfo)
        self.__add_to_metainfo(metainfo, Metainfo.NAME, name, StringValue, required=True)
        self.__add_to_metainfo(metainfo, BioMetaKeys.ORGANISM, organism, StringValue)
        self.__add_to_metainfo(metainfo, BioMetaKeys.STRAIN, strain, StringValue)
        self.__add_to_metainfo(metainfo, BioMetaKeys.REFERENCE_GENOME, reference_genome, FileReference)
        self.__add_to_metainfo(metainfo, BioMetaKeys.DATA_LINK, url, ExternalLink, required=True)
        return self.__invoke_loader(parent, 'genomeAnnotations', metainfo)

    def create_codon_table(self, parent=None, metainfo=None):
        metainfo = DataImporter._copy_metainfo(metainfo)
        return self.__invoke_loader(parent, 'codonTables', metainfo)

    def create_dbnsfp(self, parent=None, url=None, name=None, organism=None, metainfo=None):
        """
        Create a Genestack Variation Database file. ``name`` and ``url`` are required fields.
        They can be specified through the arguments or
        via a :py:class:`~odm_sdk.Metainfo` instance.

        :param parent: accession of parent folder
            (if not provided, files will be created in the ``Imported files`` folder)
        :type parent: str
        :param url: URL or local path
        :type url: str
        :param name: name of the file
        :type name: str
        :param organism: organism
        :type organism: str
        :param metainfo: metainfo object
        :type metainfo: Metainfo
        :return: file accession
        :rtype: str
        """
        metainfo = DataImporter._copy_metainfo(metainfo)
        metainfo.add_string(BioMetaKeys.DATABASE_ID, 'dbNSFP')
        self.__add_to_metainfo(metainfo, Metainfo.NAME, name, StringValue, required=True)
        self.__add_to_metainfo(metainfo, BioMetaKeys.ORGANISM, organism, StringValue)
        self.__add_to_metainfo(metainfo, BioMetaKeys.DATA_LINK, url, ExternalLink, required=True)
        return self.__invoke_loader(parent, 'dbnsfp', metainfo)

    def create_reference_genome(self,
                                parent=None,
                                name=None,
                                description='',
                                sequence_urls=None,
                                annotation_url=None,
                                organism=None,
                                assembly=None,
                                release=None,
                                strain=None,
                                metainfo=None):
        """
        Create a Genestack Reference Genome from a local or
        remote GTF or GFF
        annotation file. ``name``, ``organism`` and
        ``annotation_url`` are required fields.
        They can be specified through the arguments or
        via a :py:class:`~odm_sdk.Metainfo` instance.


        :param parent: accession of parent folder
            (if not provided, files will be created in the ``Imported files`` folder)
        :type parent: str
        :param name: name of the file
        :type name: str
        :param description: experiment description
        :type description: str
        :param sequence_urls: [OBSOLETE] list urls or local path to sequencing files
        :type sequence_urls: list
        :param annotation_url: url to annotation file
        :type annotation_url: str
        :param organism: organism
        :type organism: str
        :param assembly: assembly
        :type assembly: str
        :param release: release
        :type release: str
        :param strain: strain
        :type strain: str
        :param metainfo: metainfo object
        :type metainfo: Metainfo
        :return:
        """
        metainfo = DataImporter._copy_metainfo(metainfo)
        self.__add_to_metainfo(metainfo, Metainfo.NAME, name, StringValue, required=True)
        self.__add_to_metainfo(metainfo, BioMetaKeys.ORGANISM, organism, StringValue, required=True)
        self.__add_to_metainfo(metainfo, BioMetaKeys.STRAIN, strain, StringValue)
        self.__add_to_metainfo(metainfo, BioMetaKeys.REFERENCE_GENOME_ASSEMBLY, assembly, StringValue)
        self.__add_to_metainfo(metainfo, BioMetaKeys.REFERENCE_GENOME_RELEASE, release, StringValue)
        self.__add_to_metainfo(metainfo, ANNOTATION_KEY, annotation_url, ExternalLink, required=True)
        self.__add_to_metainfo(metainfo, metainfo.DESCRIPTION, description, StringValue)
        return self.__invoke_loader(parent, 'genomes', metainfo)

    def create_report_file(self, parent=None, name=None, urls=None, metainfo=None):
        """
        Create a Genestack Report File from a local or remote data file.
        ``name`` and ``urls`` are required fields.
        They can be specified through the arguments or
        via a :py:class:`~odm_sdk.Metainfo` instance.

        :param parent: accession of parent folder
            (if not provided, files will be created in the ``Imported files`` folder)
        :type parent: str
        :param name: name of the file
        :type name: str
        :param urls: URL or list of URLs of local file paths
        :type urls: list or str
        :param metainfo: metainfo object
        :type metainfo: Metainfo
        :return: file accession
        :rtype: str
        """
        metainfo = DataImporter._copy_metainfo(metainfo)
        self.__add_to_metainfo(metainfo, Metainfo.NAME, name, StringValue, required=True)
        self.__add_to_metainfo(metainfo, BioMetaKeys.DATA_LINK, urls, ExternalLink, required=True)
        return self.__invoke_loader(parent, 'reportFiles', metainfo)

    def create_mapped_reads_count(self,
                                  parent=None,
                                  name=None,
                                  url=None,
                                  reference_genome=None,
                                  metainfo=None):
        """
        Create a Mapped Reads Count file from a local or remote mapped reads count file.
        ``name`` and ``url`` are required fields.
        They can be specified through the arguments or
        via a :py:class:`~odm_sdk.Metainfo` instance.

        :param parent: accession of parent folder
            (if not provided, files will be created in the ``Imported files`` folder)
        :type parent: str
        :param name: name of the file
        :type name: str
        :param url: URL of a file
        :param reference_genome: reference genome accession
        :type reference_genome: str
        :param metainfo: metainfo object
        :type metainfo: Metainfo
        :return: file accession
        :rtype: str
        """
        metainfo = DataImporter._copy_metainfo(metainfo)
        self.__add_to_metainfo(metainfo, Metainfo.NAME, name, StringValue, required=True)
        self.__add_to_metainfo(metainfo, BioMetaKeys.REFERENCE_GENOME, reference_genome, FileReference)
        self.__add_to_metainfo(metainfo, BioMetaKeys.DATA_LINK, url, ExternalLink, required=True)
        return self.__invoke_loader(parent, 'mappedReadCounts', metainfo)

    def create_expression_levels(self,
                                  parent=None,
                                  unit=None,
                                  name=None,
                                  url=None,
                                  metainfo=None):
        """
        Create a Expression Levels file from a local or remote expression levels file.
        ``name``, ``url`` and ``unit`` are required fields.
        They can be specified through the arguments or
        via a :py:class:`~odm_sdk.Metainfo` instance.

        :param parent: accession of parent folder
            (if not provided, files will be created in the ``Imported files`` folder)
        :type parent: str
        :param name: name of the file
        :type name: str
        :param url: URL of the file
        :type url: str
        :param unit: unit of expression, e.g. ``TPM``, ``FPKM``
        :type unit: str
        :param metainfo: metainfo object
        :type metainfo: Metainfo
        :return: file accession
        :rtype: str
        """
        metainfo = DataImporter._copy_metainfo(metainfo)
        self.__add_to_metainfo(metainfo, Metainfo.NAME, name, StringValue, required=True)
        self.__add_to_metainfo(metainfo, BioMetaKeys.EXPRESSION_LEVEL_UNIT, unit, StringValue, required=True)
        self.__add_to_metainfo(metainfo, BioMetaKeys.DATA_LINK, url, ExternalLink, required=True)
        return self.__invoke_loader(parent, 'expressionLevels', metainfo)

    def create_gene_list(self, parent=None, name=None, url=None,
                         organism=None, metainfo=None):
        """
        Create a Gene List file from a local or remote gene list file.
        ``name``, ``url`` and ``organism`` are required fields.
        They can be specified through the arguments or
        via a :py:class:`~odm_sdk.Metainfo` instance.

        :param parent: accession of parent folder
            (if not provided, files will be created in the ``Imported files`` folder)
        :type parent: str
        :param name: name of the file
        :type name: str
        :param url: URL of a file
        :param organism: organism name
        :type organism: str
        :param metainfo: metainfo object
        :type metainfo: Metainfo
        :return: file accession
        :rtype: str
        """
        metainfo = DataImporter._copy_metainfo(metainfo)
        self.__add_to_metainfo(metainfo, Metainfo.NAME, name, StringValue, required=True)
        self.__add_to_metainfo(metainfo, BioMetaKeys.ORGANISM, organism, StringValue, required=True)
        self.__add_to_metainfo(metainfo, BioMetaKeys.DATA_LINK, url, ExternalLink, required=True)
        return self.__invoke_loader(parent, 'geneList', metainfo)

    def create_gene_expression_signature(self, parent=None, name=None, url=None,
                                         organism=None, metainfo=None):
        """
        Create a Gene Expression Signature file from a local or remote gene expression signature file.
        ``name``, ``url`` and ``organism`` are required fields.
        They can be specified through the arguments or
        via a :py:class:`~odm_sdk.Metainfo` instance.

        :param parent: accession of parent folder
            (if not provided, files will be created in the ``Imported files`` folder)
        :type parent: str
        :param name: name of the file
        :type name: str
        :param url: URL of a file
        :param organism: organism name
        :type organism: str
        :param metainfo: metainfo object
        :type metainfo: Metainfo
        :return: file accession
        :rtype: str
        """
        metainfo = DataImporter._copy_metainfo(metainfo)
        self.__add_to_metainfo(metainfo, Metainfo.NAME, name, StringValue, required=True)
        self.__add_to_metainfo(metainfo, BioMetaKeys.ORGANISM, organism, StringValue, required=True)
        self.__add_to_metainfo(metainfo, BioMetaKeys.DATA_LINK, url, ExternalLink, required=True)
        return self.__invoke_loader(parent, 'geneExpressionSignature', metainfo)

    def create_dictionary(self, parent=None, name=None, url=None, term_type=None, metainfo=None,
                          parent_dictionary=None):
        """
        Create a Dictionary file from a local or remote file.
        `owl`, `obo`, and `csv` formats are supported.
        ``name`` and ``url`` are required fields.
        They can be specified through the arguments or
        via a :py:class:`~odm_sdk.Metainfo` instance.

        :param parent: accession of parent folder
            (if not provided, files will be created in the ``Imported files`` folder)
        :type parent: str
        :param name: name of the file
        :type name: str
        :param url: URL of a file
        :type url: str
        :param term_type: dictionary term type
        :type term_type: str
        :param metainfo: metainfo object
        :type metainfo: Metainfo
        :param parent_dictionary: accession of parent dictionary
        :type parent_dictionary: str
        :return: file accession
        :rtype: str
        """

        metainfo = DataImporter._copy_metainfo(metainfo)
        self.__add_to_metainfo(metainfo, Metainfo.NAME, name, StringValue, required=True)
        self.__add_to_metainfo(metainfo, BioMetaKeys.DATA_LINK, url, ExternalLink, required=True)
        self.__add_to_metainfo(metainfo, Metainfo.PARENT_DICTIONARY, parent_dictionary, FileReference)
        self.__add_to_metainfo(metainfo, 'genestack.dictionary:termType', term_type, StringValue)
        return self.__invoke_loader(parent, 'dictionaryFiles', metainfo)

    def create_microarray_annotation(self, annotation_type, parent=None,
                                     name=None, url=None, metainfo=None):
        """
        Create a Dictionary file from a local or remote microarray annotation file.
        ``name`` and ``url`` are required fields.
        They can be specified through the arguments or
        via a :py:class:`~odm_sdk.Metainfo` instance.

        :param annotation_type: type of annotation being loaded,
            an element of :py:attr:`~odm_sdk.DataImporter.MICROARRAY_ANNOTATION_TYPES`
        :type annotation_type: str
        :param parent: accession of parent folder
            (if not provided, files will be created in the ``Imported files`` folder)
        :type parent: str
        :param name: name of the file
        :type name: str
        :param url: URL of a file
        :param metainfo: metainfo object
        :type metainfo: Metainfo
        :return: file accession
        :rtype: str
        """
        if annotation_type not in self.MICROARRAY_ANNOTATION_TYPES:
            raise GenestackException("Microarray annotation type '%s' is not "
                                     "supported, use something from "
                                     "`DataImporter.MICROARRAY_ANNOTATION_TYPES`"
                                     % annotation_type)
        metainfo = DataImporter._copy_metainfo(metainfo)
        self.__add_to_metainfo(metainfo, Metainfo.NAME, name, StringValue, required=True)
        self.__add_to_metainfo(metainfo, BioMetaKeys.DATA_LINK, url, ExternalLink, required=True)
        return self.__invoke_loader(parent, annotation_type, metainfo)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from odm_sdk import FilesUtil


class DictionarySearch(object):
    """
    Class for finding dictionaries used in templates.
    """

    def __init__(self, connection):
        self._files_util = FilesUtil(connection)

    def find_dictionary(self, path, name):
        parent = self._find_dictionary_directory(path)
        return self._files_util.find_file_by_name(name, parent=parent)

    def _find_dictionary_directory(self, path):
        head, tail = self._split_path(path)
        return self._files_util.get_folder(head, *tail)

    def _split_path(self, path):
        # search in public dictionaries as default
        if not path:
            return self._files_util.get_public_folder(), ['Dictionaries']
        # search in the home folder if not specified otherwise
        else:
            split = path.split('/')
            home_folder = self._files_util.get_home_folder()
            if split[0] in (self._files_util.get_public_folder(), home_folder):
                return split[0], split[1:]
            else:
                return home_folder, split



from odm_sdk import Application, FilesUtil, GenestackException


class CLApplication(Application):
    """
    Base class to interact with Genestack command-line applications.
    The ``APPLICATION_ID`` is mandatory. You can either pass it as an argument to the class constructor or override
    it in a child class.
    Source files and parameters are application-specific.

    """
    APPLICATION_ID = None

    def __repr__(self):
        return self.APPLICATION_ID

    def create_file(self, source_files, name=None, params=None, calculate_checksums=False, expected_checksums=None,
                    initialize=False):
        """
        Create a native Genestack file with the application and return its accession.
        If a source file is not found or is not of the expected type, an exception will be thrown.

        :param source_files: list of source files accessions
        :type source_files: list
        :param name: if a name is provided, the created file will be renamed
        :type name: str
        :param params: custom command-line arguments strings; if None,
            the application defaults will be used.
        :param params: list
        :param calculate_checksums: a flag used in the initialization script
            to compute checksums for the created files
        :type calculate_checksums: bool
        :param expected_checksums: Dict of expected checksums (``{metainfo_key: expected_checksum}``)
        :type expected_checksums: dict
        :param initialize: should initialization be started immediately
            after the file is created?
        :return: accession of created file
        :rtype: str
        """
        app_file = self.__create_file(source_files, params)

        fu = FilesUtil(self.connection)
        if name:
            fu.rename_file(app_file, name)

        if calculate_checksums:
            fu.mark_for_tests(app_file)

        if expected_checksums:
            fu.add_checksums(app_file, expected_checksums)

        if initialize:
            self.start(app_file)
        return app_file

    def __create_file(self, source_files, params=None):
        source_file_list = source_files if isinstance(source_files, list) else [source_files]
        result_file = self.invoke('createFile', source_file_list)
        if params is not None:
            self.change_command_line_arguments(result_file, params)
        return result_file

    def change_command_line_arguments(self, accession, params):
        """
        Change the command-line arguments strings in a file's metainfo.
        ``params`` is a list of command-line strings.
        Note that the syntax of command-line argument strings is application-specific.
        The only way for you to know which command-line strings to provide it is to look at
        the ``Parameters`` metainfo field of a CLA file that has the correct parameters
        specified through the graphical user interface of the application.

        If the file is not found, does not have the right file type
        or is already initialized, an exception will be thrown.

        :param accession: file accession or accession list
        :type accession: str
        :param params: list of commandlines to be set
        :type params: list
        :return: None
        """
        self.invoke('changeCommandLineArguments', self.__to_list(accession), self.__to_list(params))

    def start(self, accession):
        """
        Start file initialization.
        If the file is not found or is not of the right file type, an exception will be thrown.

        :param accession: file accession or accession list
        :type accession: str
        :return: None
        """
        self.invoke('start', self.__to_list(accession))

    def replace_file_reference(self, accession, key, accession_to_remove, accession_to_add):
        """
        Replace a file reference on the file.

        If the file is not found or is not of the right file type,
        the corresponding exceptions are thrown.
        If ``accession_to_remove`` or ``accession_to_add`` is not found,
        an exception will be thrown.

        :param accession: file accession or accession list
        :param key: key for source files
        :param accession_to_remove: accession to remove
        :param accession_to_add: accession to add
        :return: None
        """
        self.invoke('replaceFileReference', self.__to_list(accession), key, accession_to_remove, accession_to_add)

    def __to_list(self, string_or_list):
        return string_or_list if isinstance(string_or_list, list) else [string_or_list]


class IntersectApplication(CLApplication):
    """
    Parent class for all intersect applications.
    """
    INTERSECT_OTHER_SOURCE_KEY = "genestack.bio:intersect.otherSourceData"

    def create_file(self, source_files, name=None, params=None, calculate_checksums=False,
                    expected_checksums=None, initialize=False):
        """
        Same as the parent method except that intersect applications also need a separate source
        file to intersect with, so it treats the last element of the ``source_files`` array as that
        file.
        """
        if len(source_files) < 2:
            raise GenestackException('Intersect application requires at least two source files')

        main_source_files = source_files[:-1]
        other_source_file = source_files[-1]

        created_file_accession = super(IntersectApplication, self).create_file(
            main_source_files, name, params, calculate_checksums, expected_checksums, initialize
        )
        self.replace_file_reference(
            accession=created_file_accession,
            key=self.INTERSECT_OTHER_SOURCE_KEY,
            accession_to_remove=None,
            accession_to_add=other_source_file
        )
        return created_file_accession


class TestCLApplication(CLApplication):
    APPLICATION_ID = 'genestack/testcla'


# Preprocess
# preprocess unaligned reads

class TrimAdaptorsAndContaminants(CLApplication):
    APPLICATION_ID = 'genestack/fastq-mcf'


class FilterByQuality(CLApplication):
    APPLICATION_ID = 'genestack/qualityFilter'


class TrimToFixedLength(CLApplication):
    APPLICATION_ID = 'genestack/fastx-trimmer'


class SubsampleReads(CLApplication):
    APPLICATION_ID = 'genestack/subsampling'


class FilterDuplicatedReads(CLApplication):
    APPLICATION_ID = 'genestack/filter-duplicated-reads'


class TrimLowQualityBases(CLApplication):
    APPLICATION_ID = 'genestack/trim-low-quality-bases'


class MergeRawReadsApplication(Application):
    APPLICATION_ID = 'genestack/merge-raw-reads'

    def create_merged_reads(self, sources_folder, grouping_key):
        return self.invoke('createFiles', sources_folder, grouping_key)


# preprocess microarrays

class AffymetrixMicroarraysNormalizationApplication(CLApplication):
    APPLICATION_ID = 'genestack/affymetrix-normalization'


class AgilentMicroarraysNormalizationApplication(CLApplication):
    APPLICATION_ID = 'genestack/agilent-normalization'


class L1000MicroarraysNormalizationApplication(CLApplication):
    APPLICATION_ID = 'genestack/l1000-normalization'


class GenePixMicroarraysNormalizationApplication(CLApplication):
    APPLICATION_ID = 'genestack/genepix-normalization'


class InfiniumMicroarraysNormalizationApplication(CLApplication):
    APPLICATION_ID = 'genestack/infinium-methylation-normalization'


# preprocess mapped reads

class MarkDuplicated(CLApplication):
    APPLICATION_ID = 'genestack/markDuplicates'


class RemoveDuplicated(CLApplication):
    APPLICATION_ID = 'genestack/removeDuplicates'


class MergeMappedReadsApplication(CLApplication):
    APPLICATION_ID = 'genestack/merge-mapped-reads'


class AlignedReadsSubsamplingApplication(CLApplication):
    APPLICATION_ID = 'genestack/aligned-subsampling'


# preprocess variants

class VariationMergerApplication(CLApplication):
    APPLICATION_ID = 'genestack/variationMerger'


class ConcatenateVariantsApplication(CLApplication):
    APPLICATION_ID = 'genestack/concatenateVariants'


# Analise
# analyse unaligned reads

class BowtieApplication(CLApplication):
    APPLICATION_ID = 'genestack/bowtie'


class BsmapApplicationWG(CLApplication):
    APPLICATION_ID = 'genestack/bsmapWG'


class BWAApplication(CLApplication):
    APPLICATION_ID = 'genestack/bwaMapper'


class BsmapApplication(CLApplication):
    APPLICATION_ID = 'genestack/bsmap'


class TophatApplication(CLApplication):
    APPLICATION_ID = 'genestack/tophat'


# analyse mapped reads

class CuffquantApplication(CLApplication):
    APPLICATION_ID = 'genestack/cuffquant'


class MethratioApplication(CLApplication):
    APPLICATION_ID = 'genestack/methratio'


class HTSeqCountsApplication(CLApplication):
    APPLICATION_ID = 'genestack/htseqCount'


class VariationCallerApplication(CLApplication):
    APPLICATION_ID = 'genestack/variationCaller'


class VariationCaller2Application(CLApplication):
    APPLICATION_ID = 'genestack/variationCaller-v2'


class NormalizationApplication(CLApplication):
    APPLICATION_ID = 'genestack/normalization'


class IntersectGenomicFeaturesMapped(IntersectApplication):
    APPLICATION_ID = 'genestack/intersect-bam'


# analyse raw counts


class SingleCellRNASeqAnalysisApplication(CLApplication):
    APPLICATION_ID = 'genestack/rnaseq'


# analyse Differential Expression Statistics

class GOEnrichmentAnalysis(CLApplication):
    APPLICATION_ID = 'genestack/functionalEnrichmentAnalysis'


# analyse variants

class EffectPredictionApplication(CLApplication):
    APPLICATION_ID = 'genestack/snpeff'


class VariantsAssociationAnalysisApplication(CLApplication):
    APPLICATION_ID = 'genestack/variantsAssociationAnalysis'


class IntersectGenomicFeaturesVariants(IntersectApplication):
    APPLICATION_ID = 'genestack/intersect-vcf'


# Analyse microbiome content

class QiimeMicrobiomeAnalysis(CLApplication):
    APPLICATION_ID = 'genestack/qiime-report'


# Explore apps

class UnalignedReadsQC(CLApplication):
    APPLICATION_ID = 'genestack/unalignedreads-qc'


class FastQCApplicaton(CLApplication):
    APPLICATION_ID = 'genestack/fastqc-report'


class AlignedReadsQC(CLApplication):
    APPLICATION_ID = 'genestack/alignedreads-qc'


class TargetedSequencingQC(CLApplication):
    APPLICATION_ID = 'genestack/alignedreads-qc-enrichment'


class ArrayQualityMetricsApplication(CLApplication):
    APPLICATION_ID = 'genestack/arrayqualitymetrics'


class DoseResponseApplication(CLApplication):
    APPLICATION_ID = 'genestack/dose-response'


class SingleCellRNASeqVisualiserApplication(CLApplication):
    APPLICATION_ID = 'genestack/scrvis'



import cmd
import os
import shlex
import sys
from argparse import ArgumentParser, HelpFormatter
from traceback import print_exc

from odm_sdk import (GenestackAuthenticationException, GenestackException,
                              GenestackVersionException)
from .utils import get_connection, isatty, make_connection_parser, get_terminal_width
from .version import __version__

if isatty():
    # To have autocomplete and console navigation on windows you need to have pyreadline installed.
    # Pyreadline replace readline with alias to itself.
    try:
        import readline
    except:
        pass

HELP_SEPARATOR = '\n%s  Help  %s' % ('=' * 20, '=' * 20)


def get_help(parser):
    """
    Return help text for the parser.

    :param parser: parser
    :type parser: ArgumentParser
    :return: fully formatted help string
    :rtype: str
    """
    # Code almost identical with parser.print_help() with next changes:
    # it return text instead print
    # it place group 'command arguments' to the first place
    width = min(get_terminal_width(), 100)
    formatter = HelpFormatter(prog=parser.prog, max_help_position=30, width=width)
    # usage
    formatter.add_usage(parser.usage, parser._actions,
                        parser._mutually_exclusive_groups)
    # description
    formatter.add_text(parser.description)

    # positionals, optionals and user-defined groups
    groups = list(parser._action_groups)
    groups.sort(key=lambda x: x.title != 'command arguments')
    for action_group in groups:
        formatter.start_section(action_group.title)
        formatter.add_text(action_group.description)
        formatter.add_arguments(action_group._group_actions)
        formatter.end_section()

    # epilog
    formatter.add_text(parser.epilog)

    # determine help from format above
    return formatter.format_help()


class Command(object):
    """
    Command class to be inherited.

        - ``COMMAND``: name of the command
        - ``DESCRIPTION``: description as shown in the help message
        - ``OFFLINE``: set to ``True`` if the command does not require a connection to the Genestack server.
    """
    COMMAND = None
    DESCRIPTION = ''
    OFFLINE = False

    # TODO make private all that possible

    def __init__(self):
        self.connection = None
        self.args = None

    def get_command_parser(self, parser=None):
        """
        Returns a command parser. This function is called each time before a command is executed.
        To add new arguments to the command,
        you should override the :py:meth:`~odm_sdk.shell.Command.update_parser` method.

        :param parser: base argument parser. For offline commands and commands inside shell, it will be ``None``.
            For the other cases, it will be the result of :py:func:`~odm_sdk.make_connection_parser`
        :type parser: argparse.ArgumentParser
        :return: parser
        :rtype: argparse.ArgumentParser
        """
        parser = parser or ArgumentParser(description=self.DESCRIPTION)
        parser.prog = '%s %s' % (parser.prog, self.COMMAND)
        parser.description = self.DESCRIPTION
        group = parser.add_argument_group("command arguments")
        self.update_parser(group)
        return parser

    def update_parser(self, parent):
        """
        Add arguments for the command. Should be overridden in child classes.

        :param parent: argument group
        :type parent: argparse._ArgumentGroup
        :rtype: None
        """
        pass

    def set_connection(self, conn):
        """
        Set a connection for the command.

        :param conn: connection
        :type conn: odm_sdk.Connection
        """
        self.connection = conn

    def set_arguments(self, args):
        """
        Set parsed arguments for the command.

        :param args: parsed arguments
        :type args: argparse.Namespace
        """
        self.args = args

    def get_short_description(self):
        """
        Returns a short description for the command. Used in the "help" message.

        :return short description
        :rtype: str
        """
        return self.DESCRIPTION

    def run(self):
        """
        Override this method to implement the command action.

        Return value of this method is always ignored.
        If this method raises an exception, the command will be treated as failed.

        If this command is executed in the shell mode,
        the failed state is ignored, otherwise exit code ``1`` is returned.

        Raise :py:class:`~odm_sdk.exceptions.GenestackException` to indicate command failure
        without showing the stacktrace.

        :rtype: None
        """
        raise NotImplementedError()


class GenestackShell(cmd.Cmd):
    """
    Arguments to be overridden in children:

        - ``INTRO``: greeting at start of shell mode
        - ``COMMAND_LIST``: list of available commands
        - ``DESCRIPTION``: description for help.

    Run as script:

        .. code-block:: sh

            script.py [connection_args] command [command_args]

    Run as shell:

        .. code-block:: sh

            script.py [connection_args]


    Default shell commands:
        - ``help``: show help about shell or command
        - ``quit``: quits shell
        - ``ctrl+D``: quits shell
    """
    INTRO = ''
    COMMAND_LIST = []
    DESCRIPTION = "Shell and commandline application"

    COMMANDS = {}  # filled in init

    # Don't add docstrings to methods, all methods should be considered private, but due design of Cmd it is not possible
    # TODO make private all that possible

    def get_history_file_path(self):
        """
        Get path to history file.

        :return: path to history file
        :rtype: str
        """
        return os.path.join(os.path.expanduser("~"), '.%s' % self.__class__.__name__)

    def get_names(self):
        return [x for x in cmd.Cmd.get_names(self) if x != 'do_EOF']

    def __init__(self, *args, **kwargs):
        self.COMMANDS = {command.COMMAND: command for command in self.COMMAND_LIST}
        cmd.Cmd.__init__(self, *args, **kwargs)
        self.connection = None

    def get_shell_parser(self, offline=False):
        """
        Returns the parser for shell arguments.

        :return: parser for shell commands
        :rtype: argparse.ArgumentParser
        """
        parents = [] if offline else [make_connection_parser()]
        parser = ArgumentParser(conflict_handler='resolve', description=self.DESCRIPTION, parents=parents)

        # override default help
        parser.add_argument('-h', '--help', action='store_true', help="show this help message and exit")
        parser.add_argument('-v', '--version', action='store_true', help="show version")
        parser.add_argument('command', metavar='<command>', help='"%s" or empty to use shell' % '", "'.join(self.COMMANDS), nargs='?')
        return parser

    def setup_connection(self, args=None):
        try:
            self.connection = get_connection(args)
        except GenestackVersionException as e:
            sys.stderr.write(str(e))
            sys.stderr.write('\n')
            exit(13)

    def preloop(self):
        # Entry point. Check whether we should run a script and exit, or start an interactive shell.

        parser = self.get_shell_parser()
        args, others = parser.parse_known_args()

        if args.version:
            print(__version__)
            exit(0)

        command = self.COMMANDS.get(args.command)
        if command:
            command = command()
        elif args.command:
            print("*** Unknown command: %s" % args.command)
            print(get_help(parser))
            exit(0)
        elif others:
            print("*** Unknown arguments: %s" % ' '.join(others))
            print(get_help(parser))
            exit(0)

        if args.help:
            if not command:
                print(get_help(parser))
            elif command.OFFLINE:
                print(get_help(command.get_command_parser()))
            else:
                print(get_help(command.get_command_parser(make_connection_parser())))
            exit(0)

        if command:
            if not command.OFFLINE:
                self.setup_connection(args)
            else:
                # parse arguments that have same name as connection parser
                parser = self.get_shell_parser(offline=True)
                _, others = parser.parse_known_args()

            exit_code = self.process_command(command, others)
            exit(exit_code)

        # do shell
        try:
            readline.read_history_file(self.get_history_file_path())
            readline.set_history_length(1000)
        except (IOError, NameError):
            pass
        self.set_shell_user(args)

    def set_shell_user(self, args):
        """
        Set the connection for shell mode.

        :param args: script arguments
        :type args: argparse.Namespace
        """
        # set user for shell
        self.setup_connection(args)
        try:
            email = self.connection.whoami()
            self.prompt = '%s> ' % email
        except GenestackAuthenticationException:
            self.prompt = 'anonymous>'
        if self.connection.debug:
            debug_string = ' (with debug enabled)'
        else:
            debug_string = ''
        self.intro = ('odm_sdk v{version}{debug_string}\n'
                      '{intro}'.format(version=__version__,
                                       debug_string=debug_string,
                                       intro=self.INTRO)
                      )

    def postloop(self):
        try:
            readline.write_history_file(self.get_history_file_path())
        except (IOError, NameError):
            pass

    def do_EOF(self, line):
        return True

    do_quit = do_EOF

    def process_command(self, command, argument_list, shell=False):
        """
        Runs the given command with the provided arguments and returns the exit code

        :param command: command
        :type command: Command
        :param argument_list: the list of arguments for the command
        :type argument_list: list
        :param shell: should we use shell mode?
        :type shell: bool
        :return: 0 if the command was executed successfully, 1 otherwise

        :rtype: int
        """
        if shell or command.OFFLINE:
            p = command.get_command_parser()
        else:
            p = command.get_command_parser(make_connection_parser())
        try:
            args = p.parse_args(argument_list)
        except SystemExit:
            return 1
        command.set_connection(self.connection)
        command.set_arguments(args)
        try:
            command.run()
            return 0
        except (KeyboardInterrupt, EOFError):
            print()
            print("Command interrupted.")
        except GenestackException as e:
            sys.stdout.flush()
            sys.stderr.write('%s\n' % e)
        except Exception:
            sys.stdout.flush()
            print_exc()
        return 1

    def do_debug(self, line):
        self.connection.debug = not self.connection.debug
        if self.connection.debug:
            print('Debug enabled')
        else:
            print('Debug disabled')

    def get_commands_for_help(self):
        """
        Return list of command - description pairs to shown in shell help command.

        :return: command - description pairs
        :rtype: list[(str, str)]
        """
        commands = [('quit', 'Exit shell.'), ('debug', 'Toggle debug for connection.')]
        for name, value in self.COMMANDS.items():
            commands.append((name, value().get_short_description()))
        return sorted(commands)

    def do_help(self, line):
        print()

        command = self.COMMANDS.get(line)
        if command:
            print(get_help(command().get_command_parser()))
            return

        if not line:
            print(self.doc_header)
            print('=' * len(self.doc_header))
            commands = self.get_commands_for_help()
            max_size = max(len(command_name) for command_name, _ in commands)
            help_size = max((20, max_size + 3))
            for command_name, short_description in self.get_commands_for_help():
                print('%-*s%s' % (help_size, command_name, short_description))
            print('=' * len(self.doc_header))

        try:
            getattr(self, 'help_' + line)()
            return
        except AttributeError:
            pass

        try:
            doc = getattr(self, 'do_' + line).__doc__
            if doc:
                self.stdout.write('%s\n' % str(doc))
                return
        except AttributeError:
            pass

        self.stdout.write('%s\n' % str(self.nohelp % (line,)))

    def emptyline(self):
        # override default cmd behavior.
        pass

    def default(self, line):
        try:
            args = shlex.split(line)
        except Exception as e:
            sys.stderr.write(str(e))
            sys.stderr.write('\n')
            return
        if args and args[0] in self.COMMANDS:
            self.process_command(self.COMMANDS[args[0]](), args[1:], shell=True)
        else:
            self.stdout.write('*** Unknown command: %s\n' % line)

    def completenames(self, text, *ignored):
        dotext = 'do_' + text
        commands = [a[3:] for a in self.get_names() if a.startswith(dotext)]
        commands += [x for x in self.COMMANDS.keys() if x.startswith(text)]
        return commands

    def cmdloop(self, intro=None):
        # TODO on ctrl+C don't exit
        try:
            cmd.Cmd.cmdloop(self, intro=intro)
        except (KeyboardInterrupt, EOFError):
            print()
            self.postloop()



from urllib.error import URLError

MASTER_BRANCH = 'https://github.com/genestack/python-client/archive/master.zip'
PYPI_PACKAGE = 'genestack-client'


class GenestackBaseException(Exception):
    """
    Base class for Genestack exceptions.

    Use it to catch all exceptions raised explicitly by Genestack Python Client.
    """
    pass


class GenestackException(GenestackBaseException):
    """
    Client-side  exception class.

    Raise its instances (instead of :py:class:`~exceptions.Exception`)
    if anything is wrong on client side.
    """
    pass


class GenestackServerException(GenestackException):
    """
    Server-side exception class.

    Raised when Genestack server returns an error response
    (error message generated by Genestack Java code, not an HTTP error).
    """

    def __init__(self, message, path, post_data, debug=False, stack_trace=None):
        """
        :param message: exception message
        :type message: str
        :param path: path after server URL of connection.
        :type path: str
        :param post_data: POST data (file or dict)
        :type debug: bool
        :param debug: flag if stack trace should be printed
        :param stack_trace: server stack trace
        :type stack_trace: str
        """
        message = (message.decode('utf-8', 'ignore')
                   if isinstance(message, bytes) else message)

        GenestackException.__init__(self, message, path, post_data, debug, stack_trace)
        self.message = message
        self.debug = debug
        self.stack_trace = stack_trace
        self.path = path
        self.post_data = post_data

    def __str__(self):
        if isinstance(self.post_data, dict):
            message = 'Got error "%s" at call of method "%s" of "%s"' % (
                self.message,
                self.post_data.get('method', '<unknown>'),
                self.path
            )
        else:
            # upload file
            message = 'Got error "%s" at call of "%s"' % (
                self.message,
                self.path
            )
        if self.stack_trace:
            if self.debug:
                message += '\nStacktrace from server is:\n%s' % self.stack_trace
            else:
                message += '\nEnable debug option to retrieve traceback'

        return message


class GenestackResponseError(GenestackBaseException, URLError):
    """
    Wrapper for HTTP response errors.

    Extends :py:class:`urllib2.URLError` for backward compatibility.
    """
    def __init__(self, reason):
        self.args = reason,
        self.reason = reason

    def __str__(self):
        return '<urlopen error %s>' % self.reason


class GenestackConnectionFailure(GenestackBaseException, URLError):
    """
    Wrapper for server connection failures.

    Extends :py:class:`urllib2.URLError` for backward compatibility.
    """
    def __init__(self, message):
        self.message = "<connection failed %s>" % message

    def __str__(self):
        return self.message


class GenestackAuthenticationException(GenestackException):
    """
    Exception thrown on an authentication error response from server.
    """
    pass


class GenestackVersionException(GenestackException):
    """
    Exception thrown if server requires a newer version on Python Client.
    """

    def __init__(self, current_version, required_version=None):
        """

        :param current_version: current version
        :type current_version: str
        :param required_version: minimum required version
        :type required_version: str
        """
        if not required_version:
            message = 'Cannot get required version from server.\n'

        super(GenestackVersionException, self).__init__(message)



from setuptools import setup, find_packages
import re

regex = re.compile(r'^--.*$')
# Read requirements file to download deps with filtering params.
with open('requirements.txt') as f:
    required = [line for line in f.read().splitlines() if not regex.match(line)]

with open('requirements-internal.txt') as f:
    required_internal = [line for line in f.read().splitlines() if not regex.match(line)]

# Import version directly, without execute __init__.py script
exec(open('odm_sdk/version.py').read())

setup(
    name='odm-sdk',
    install_requires=required + required_internal,
    version=__version__,
    packages=find_packages(),
    url='https://github.com/genestack/sdk',
    license='MIT',
    author='Genestack Limited',
    author_email='',
    description='SDK for interacting with the Open Data Manager',
    long_description=__doc__,
    long_description_content_type="text/markdown",
    keywords=['genestack', 'odm', 'import', 'share', 'create',
              'delete', 'curate', 'genomics', 'api'],
    include_package_data=True,
    python_requires='>=3.10',
    entry_points={
        'console_scripts': [
            'odm-user-setup = odm_sdk.scripts.user_setup:main',
            'odm-shell = odm_sdk.scripts.shell:main',
            'odm-import-data = odm_sdk.scripts.import_ODM_data:main',
            'odm-create-users = odm_sdk.scripts.users.create_users:main',
            'odm-update-template = odm_sdk.scripts.metainfo_templates.update_template:main',
            'odm-delete-template = odm_sdk.scripts.delete_study_or_template:main',
            'odm-update-dictionary = odm_sdk.scripts.dictionaries.load_init_share_dictionaries:main',
            'odm-curate-study = odm_sdk.scripts.study_management.create_curation_file:main',
            'odm-delete-study = odm_sdk.scripts.delete_study_or_template:main',
            'odm-share-study = odm_sdk.scripts.study_management.share_study_with_group:main',
            'odm-geo-prepare = odm_sdk.scripts.study_management.GEO_prepare:main',
            'odm-upload-reference-genome = odm_sdk.scripts.upload_reference_genome:main',
        ],
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        'Topic :: Software Development',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Software Development :: Libraries :: Python Modules',
        ],

    project_urls={
        'Documentation': 'https://genestack-client.readthedocs.io/en/stable/',
        'Source': 'https://github.com/genestack/python-client/',
        'Tracker': 'https://github.com/genestack/python-client/issues',
    },
)



import argparse
import json
import shlex
import sys
from datetime import datetime

from odm_sdk.shell import Command, GenestackShell
from odm_sdk import ShareUtil

APPLICATION_SHELL = 'genestack/shell'


def serialize_params(params):
    args = []
    for arg in params:
        try:
            args.append(json.loads(arg))
        except ValueError:
            args.append(arg)
    return args


class Call(Command):
    COMMAND = 'call'
    DESCRIPTION = 'call another application\'s method'
    OFFLINE = False

    def update_parser(self, p):
        p.add_argument(
            'applicationId',
            help='full application id'
        )
        p.add_argument(
            'method',
            help='application method'
        )
        p.add_argument(
            'params', nargs=argparse.REMAINDER,
            help='params'
        )

    def do_request(self):
        params = serialize_params(self.args.params)
        return self.connection.application(self.args.applicationId).invoke(self.args.method, *params)

    def run(self):
        res = self.do_request()
        print(json.dumps(res, indent=2))


class Time(Call):
    DESCRIPTION = 'invoke with timer'
    COMMAND = 'time'

    def run(self):
        start = datetime.now()
        Call.run(self)
        print('Execution time: %s' % (datetime.now() - start))


class Groups(Command):
    DESCRIPTION = 'print information about user groups'
    COMMAND = 'groups'

    def run(self):
        share_util = ShareUtil(self.connection)
        print('User groups:')
        for accession, name in share_util.get_available_sharing_groups().items():
            print('  %s (%s)' % (name, accession))


class Shell(GenestackShell):
    COMMAND_LIST = [Time, Call, Groups]

    def get_commands_for_help(self):
        commands = GenestackShell.get_commands_for_help(self)
        for data in self.connection.application(APPLICATION_SHELL).invoke('getCommands'):
            commands.append((data['name'],
                            'args: %s returns: %s. %s' % (
                                                ', '.join(data['params']) or 'None',
                                                data['returns'],
                                                data['docs'] or 'No documentation')))
        return commands

    def default(self, line):
        try:
            args = shlex.split(line)
        except Exception as e:
            sys.stderr.write(str(e))
            sys.stderr.write('\n')
            return
        if args and args[0] in self.COMMANDS:
            self.process_command(self.COMMANDS[args[0]](), args[1:], self.connection)
        else:
            self.process_command(Call(), ['genestack/shell'] + args, self.connection)


def main():
    shell = Shell()
    shell.cmdloop()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# coding=utf-8

"""
Local script used to perform automated curation.
"""
from __future__ import print_function

import json
import sys

from odm_sdk import Application, FilesUtil, make_connection_parser, get_connection
from jsonschema import validate, ValidationError

RULES_SCHEMA = {
    'type': 'array',
    'items': {
        'type': 'object',
        'properties': {
            'object_type': {'type': 'string',
                            'enum': ['sample', 'study',
                                     'library', 'preparation',
                                     'expression', 'variant', 'flow_cytometry']},
            'genestack_key': {
                'type': ['string', 'array'],
                # if `genestack_key` contains an array, it must be an array of 2 strings
                # (value key + unit key) for use in KeyWithUnitMapper
                'minItems': 2,
                'maxItems': 2,
                'items': {'type': 'string'}
            },
            'raw_keys': {
                'type': 'array',
                'items': {'type': 'string'}
            },
            'dictionary': {'type': 'string'},
            'rules': {
                'type': 'object',
                'additionalProperties': {
                    'type': ['array', 'string', 'null']
                }
            },
            'type': {
                'type': 'string'
            }
        },
        'additionalProperties': False,
        'required': ['genestack_key', 'object_type', 'raw_keys']
    }
}

DEFAULT_VERSION_MESSAGE = "Edited using curation script"


class CurationApplication(Application):
    APPLICATION_ID = "genestack/curation"

    def create_files(self, target_studies_accessions, overwrite=False,
                     dry_run=False, rules=None, new_version_message=None, publish_new_version=True):
        return self.invoke('createCurationFiles', target_studies_accessions,
                           overwrite, dry_run, rules, new_version_message, publish_new_version)


def main():
    parser = make_connection_parser()
    parser.add_argument('accessions', metavar='<accessions>', nargs='*',
                        help='Accession(s) of the study (or studies) to curate')
    parser.add_argument('--dry-run', action='store_true', help='Do not write any metainfo')
    parser.add_argument('--overwrite', action='store_true', help='Overwrite existing curated keys')
    parser.add_argument('--rules', metavar='<rules_file>',
                        help='A JSON-formatted curation rules file')
    # make `publish_new_version = True` a default, allow overriding
    version_args = parser.add_mutually_exclusive_group()
    version_args.add_argument('--do-not-publish', action='store_false', dest='publish_new_version',
                              help='Do not publish any version, save changes to the draft version.')
    version_args.add_argument('--version-message', dest='new_version_message',
                              metavar='<new_version_message>', default=DEFAULT_VERSION_MESSAGE,
                              help='Set a version message to explain the reason for update. If not '
                                   'specified, the default message "{}" will be used.'.format(
                                  DEFAULT_VERSION_MESSAGE))

    arguments = parser.parse_args()

    if not arguments.accessions:
        print("You must specify a target accession(s) for curation")
        sys.exit(1)

    if arguments.rules is None:
        print("You must specify a path to curation rules file")
        sys.exit(1)

    if not arguments.new_version_message:
        print("New version message must not be blank if specified")
        sys.exit(1)

    print("Connecting to Genestack...")
    connection = get_connection(arguments)
    curation_app = CurationApplication(connection)
    files_util = FilesUtil(connection)

    with open(arguments.rules, 'r') as rules_file:
        rules = json.load(rules_file)
    try:
        validate(rules, RULES_SCHEMA)
    except ValidationError as e:
        print("Invalid JSON curation file: " + e.message)
        sys.exit(1)

    print("Creating and initializing automated curation file...")

    accessions = curation_app.create_files(arguments.accessions,
                                           arguments.overwrite,
                                           arguments.dry_run,
                                           rules,
                                           arguments.new_version_message,
                                           arguments.publish_new_version
                                           )

    for accession in accessions:
        print("Automated curation task started! (%s)" % accession)


if __name__ == "__main__":
    main()



class BioMetaKeys(object):
    """
    Collection of bio constants.
    """
    METHOD = 'genestack.bio:method'
    ORGANISM = 'genestack.bio:organism'
    STRAIN = 'genestack.bio:strain-breed-cultivar'
    BREED = STRAIN
    CULTIVAR = STRAIN
    TISSUE = 'genestack.bio:tissue'
    CELL_TYPE = 'genestack.bio:cellType'
    CELL_LINE = 'genestack.bio:cellLine'
    DEVELOPMENTAL_STAGE = 'genestack.bio:developmentalStage'
    DISEASE_STATE = 'genestack.bio:disease'
    DISEASE_STAGE = 'genestack.bio:diseaseStage'
    COMPOUND = 'genestack.bio:compoundTreatment/compound'
    COMPOUND_DOSE_VALUE = 'genestack.bio:compoundTreatment/dose/value'
    COMPOUND_DOSE_UNIT = 'genestack.bio:compoundTreatment/dose/unit'
    COMPOUND_TIME_VALUE = 'genestack.bio:compoundTreatment/time/value'
    COMPOUND_TIME_UNIT = 'genestack.bio:compoundTreatment/time/unit'
    AGE = 'genestack.bio:age'
    SEX = 'genestack.bio:sex'
    HUMAN_ETHNIC_GROUP = 'genestack.bio:humanEthnicGroup'
    EXPRESSION_LEVEL_UNIT = 'Expression level unit'
    READS_LINK = 'genestack.url:reads'
    DATA_LINK = 'genestack.url:data'
    BAM_FILE_LINK = 'genestack.url:bamfile'
    REFERENCE_SEQUENCES = 'genestack.bio:referenceSequences'
    REFERENCE_GENOME = 'genestack.bio:referenceGenome'
    REFERENCE_GENOME_ASSEMBLY = 'genestack.bio:referenceGenomeAssembly'
    REFERENCE_GENOME_RELEASE = 'genestack.bio:referenceGenomeRelease'
    STUDY_SHORT_NAME = 'genestack.bio:shortName'
    DATABASE_ID = 'genestack.bio:databaseId'

    RNASEQ_TECHNOLOGY = 'Expression profiling by high throughput sequencing'
    DNASEQ_TECHNOLOGY = 'Genome variation profiling by high throughput sequencing'
    CHIPSEQ_TECHNOLOGY = 'Genome binding/occupancy profiling by high throughput sequencing'
    MICROARRAY_TECHNOLOGY = 'Expression profiling by array'
    DNA_MICROARRAY_TECHNOLOGY = 'Genome variation profiling by SNP array'
    # SAMPLE_ID = 'genestack.bio:sampleId'      # still used in 1000 loader.
    EXTRACTED_MOLECULE = 'genestack.bio:extractedMolecule'
    PLATFORM = 'genestack.bio:platform'
    SOURCE_DATA_PREFIX = 'sourceData:'
    SECONDARY_ACCESSION = 'genestack.bio:secondaryAccession'



import datetime
import os
import sys
from pprint import pformat
from urllib.parse import unquote, urlparse

from odm_sdk import GenestackException


class MetainfoScalarValue(dict):
    _TYPE = None

    def _set_fields(self, value, stringify=True):
        if isinstance(value, dict):
            val = value.copy()
            val['type'] = self._TYPE
            self.update(val)
        else:
            self.update({'type': self._TYPE, 'value': value})

    def __init__(self, value):
        super(MetainfoScalarValue, self).__init__()
        self._set_fields(value)

    def __repr__(self):
        return '{}({!r})'.format(self.__class__.__name__, self.get('value'))

    def _print_deprecation_message(self):
        sys.stderr.write(
            "Metainfo type `%s` is deprecated. Use a combination of other metainfo types to store "
            "a complex value.\n" % self._TYPE
        )


class StringValue(MetainfoScalarValue):
    _TYPE = 'string'

    def get_string(self):
        return self.get('value')


class BooleanValue(MetainfoScalarValue):
    _TYPE = 'boolean'

    def get_boolean(self):
        return bool(self.get('value'))

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, self.get_boolean())


class IntegerValue(MetainfoScalarValue):
    _TYPE = 'integer'

    def get_int(self):
        return int(self.get('value'))

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, self.get_int())


class MemorySizeValue(MetainfoScalarValue):
    _TYPE = 'memorySize'

    def get_int(self):
        return int(self.get('value'))

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, self.get_int())


class DecimalValue(MetainfoScalarValue):
    _TYPE = 'decimal'

    def get_decimal(self):
        return float(self.get('value'))

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, self.get_decimal())


class ExternalLink(MetainfoScalarValue):
    _TYPE = 'externalLink'

    def __init__(self, url, text=None, fmt=None):
        super(MetainfoScalarValue, self).__init__()
        if not text:
            text = os.path.basename(urlparse(unquote(url)).path)
        self._set_fields({'text': text, 'url': url, 'format': fmt})

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, self.get_url())

    def get_text(self):
        return self.get('text')

    def get_url(self):
        return self.get('url')

    def get_format(self):
        return self.get('format')


class FileReference(MetainfoScalarValue):
    _TYPE = 'file'

    def __init__(self, accession):
        super(MetainfoScalarValue, self).__init__()
        self._set_fields({'accession': accession})

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, self.get_accession())

    def get_accession(self):
        return self.get('accession')


class DateTimeValue(MetainfoScalarValue):
    _TYPE = 'datetime'

    _DATE_TIME_FORMAT = '%Y-%m-%d %H:%M:%S'
    _DATE_FORMAT = '%Y-%m-%d'

    def __init__(self, time):
        super(MetainfoScalarValue, self).__init__()
        milliseconds = self._parse_date_time(time)
        self._set_fields({'date': milliseconds})

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, self.get_date())

    @staticmethod
    def _can_be_cast_to_int(time_str):
        try:
            int(time_str)
            return True
        except ValueError:
            return False

    @classmethod
    def _parse_date_time(cls, time):
        if isinstance(time, str):
            if cls._can_be_cast_to_int(time):
                return int(time)
            try:
                time = datetime.datetime.strptime(time, cls._DATE_TIME_FORMAT)
            except ValueError:
                try:
                    time = datetime.datetime.strptime(time, cls._DATE_FORMAT)
                except ValueError:
                    raise GenestackException('Unexpected datetime string format: %s, '
                                             'specify date in on of the next format: "%s", "%s"' % (time,
                                                                                                    cls._DATE_TIME_FORMAT,
                                                                                                    cls._DATE_FORMAT))
        if isinstance(time, datetime.datetime):
            diff = time - datetime.datetime(1970, 1, 1)
            milliseconds = (diff.days * 24 * 60 * 60 + diff.seconds) * 1000 + (diff.microseconds / 1000)
        elif isinstance(time, datetime.date):
            diff = time - datetime.date(1970, 1, 1)
            milliseconds = diff.days * 24 * 60 * 60 * 1000
        elif isinstance(time, float):
            milliseconds = int(time * 1000)
        else:
            raise GenestackException('Unexpected datetime input type: %s' % type(time))
        return milliseconds

    def get_milliseconds(self):
        return float(self.get('date'))

    def get_date(self):
        return datetime.datetime.fromtimestamp(self.get_milliseconds()/1000.0)


class Person(MetainfoScalarValue):
    _TYPE = 'person'

    def __init__(self, name, phone=None, email=None):
        super(MetainfoScalarValue, self).__init__()
        self._print_deprecation_message()
        self._set_fields({'name': name, 'phone': phone, 'email': email})

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__,
                               pformat(self.get_person()))

    def get_person(self):
        return {key: self.get(key) for key in {'name', 'phone', 'email'}}


class Publication(MetainfoScalarValue):
    _TYPE = 'publication'

    def __init__(self, title, authors, journal_name, issue_date, identifiers=None, issue_number=None, pages=None):
        super(MetainfoScalarValue, self).__init__()
        self._print_deprecation_message()
        self._set_fields({
            'identifiers': identifiers if identifiers else {},
            'journalName': journal_name,
            'issueDate': issue_date,
            'title': title,
            'authors': authors,
            'issueNumber': issue_number,
            'pages': pages
        })

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__,
                               pformat(self.get_publication()))

    def get_publication(self):
        return {key: self.get(key) for key in {'identifiers', 'journalName', 'issueDate', 'title', 'authors',
                                               'issueNumber', 'pages'}}


class Organization(MetainfoScalarValue):
    _TYPE = 'organization'

    def __init__(self, name, department=None, country=None, city=None, street=None,
                 postal_code=None, state=None, phone=None, email=None, url=None):
        super(MetainfoScalarValue, self).__init__()
        self._print_deprecation_message()
        self._set_fields({
            'name': name,
            'department': department,
            'country': country,
            'city': city,
            'street': street,
            'postalCode': postal_code,
            'state': state,
            'phone': phone,
            'email': email,
            'url': url
        })

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__,
                               pformat(self.get_organization()))

    def get_organization(self):
        return {key: self.get(key) for key in {'name', 'department', 'country', 'city', 'street',
                                               'postalCode', 'state', 'phone', 'email', 'url'}}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# This script adds existing users to existing groups.
# It requires tsv file groups_and_users.tsv to be present in the same directory as this script.
# groups_and_users.tsv should contain lines of the next format: groups, separated by commas,
# and users, separated by commas.
# Groups and users should be separated by tabs.

# Same groups may be repeated several times on different lines.

# For example:
# group1,group2 alice@alphacorp.com,suzy@alphacorp.com
# group3  bob@alphacorp.com

# See the full instruction at:
# https://genestack.atlassian.net/wiki/spaces/ODMP/pages/1143242753/How+to+create+groups+and+add+existing+users+to+them+via+python+scripts
from __future__ import print_function, unicode_literals


import sys

from odm_sdk.utils import make_connection_parser, get_connection
from odm_sdk.scripts.utils import colored, TsvReader, GREEN, BLUE, RED


def get_groups(connection):
    return connection.application('genestack/groupsadmin-api').invoke('getData')['groups']


def add_group_member(connection, group_accession, email):
    return connection.application('genestack/groupsadmin-api').invoke(
        'addMemberToGroup', group_accession, email)


def add_user_to_groups(connection, user, non_existing_groups,
                       existing_group_names_and_user_emails, group_names_and_accessions):
    email = user.get("email")
    groups = user.get("groups")
    for group in groups:
        if group not in non_existing_groups:
            if email not in existing_group_names_and_user_emails[group]:
                add_group_member(connection, group_names_and_accessions[group], email)
                print(colored("Added user %s to group %s" % (email, group),
                              GREEN))
            else:
                print(colored("User %s already belongs to group %s" % (email, group),
                              BLUE))


def read_data_for_adding_users(args, unique_group_names):
    all_users_to_be_added = []
    incorrect_rows = []
    rows_number = 0
    with TsvReader(args.file_with_users) as usersfile:
        for row in usersfile:
            rows_number += 1
            try:
                groups, emails = row[0].strip(), row[1].strip()
                if len(groups) == 0 or len(emails) == 0:
                    raise IndexError
            except IndexError:
                incorrect_rows.append(row)
                continue
            user_groups = [group_name.strip() for group_name in groups.split(",") if group_name]
            unique_group_names.update(user_groups)

            all_users_to_be_added += [
                {'email': email.strip(), 'groups': user_groups} for email in emails.split(",")
                if email
            ]
        incorrect_rows_number = len(incorrect_rows)
        if incorrect_rows_number > 0:
            print(colored("Error: each line should contain groups separated by comma and users' "
                          "emails separated by comma, groups and users should be separated by tab.",
                          RED))
            if rows_number != incorrect_rows_number:
                print(colored("Lines to correct:", RED))
                for incorrect_row in incorrect_rows:
                    print(colored(incorrect_row, RED))
            print(colored("FAILED: Correct data and rerun this script.", RED))
            sys.exit()
    return all_users_to_be_added


def check_all_users_exist(connection, all_users_to_be_added):
    existing_users = connection.application('genestack/usersadmin-api').invoke('listUsers')
    existing_emails = {existing_user['email'] for existing_user in existing_users}
    non_existing_emails = {
        user['email'] for user in all_users_to_be_added
        if user['email'] not in existing_emails
    }
    return non_existing_emails


def check_all_groups_exist(connection, group_names_and_accessions,
                           existing_group_names_and_user_emails, unique_group_names):
    all_groups = get_groups(connection)
    for group in all_groups:
        group_names_and_accessions[group['name']] = group['accession']
        members_emails = [member['email'] for member in group['members']]
        existing_group_names_and_user_emails[group['name']] = members_emails
    return unique_group_names - set(group_names_and_accessions.keys())


def main():
    parser = make_connection_parser()
    parser.add_argument('--file_with_users', metavar='<file_with_users>',
                        default="groups_and_users.tsv",
                        help='add existing users to existing groups')
    connection = get_connection()
    args = parser.parse_args()

    unique_group_names = set()

    group_names_and_accessions = {}
    existing_group_names_and_user_emails = {}

    all_users_to_be_added = read_data_for_adding_users(args, unique_group_names)

    non_existing_emails = check_all_users_exist(connection, all_users_to_be_added)
    all_users_exist = len(non_existing_emails) == 0

    non_existing_groups = check_all_groups_exist(connection, group_names_and_accessions,
                                                 existing_group_names_and_user_emails,
                                                 unique_group_names)
    all_groups_exist = len(non_existing_groups) == 0

    for user in all_users_to_be_added:
        if user['email'] not in non_existing_emails:
            add_user_to_groups(connection, user, non_existing_groups,
                               existing_group_names_and_user_emails, group_names_and_accessions)

    if all_groups_exist and all_users_exist:
        print(colored("Success", GREEN))
    else:
        if not all_groups_exist:
            print(colored("Groups %s don't exist" % non_existing_groups, RED))
        if not all_users_exist:
            print(colored("Users %s don't exist" % non_existing_emails, RED))
        if not all_users_exist or not all_groups_exist:
            print(colored("Create missed items and rerun this script with the same input file",
                          RED))


if __name__ == "__main__":
    main()



import pandas as pd
import os
import csv
import sys
import re


def get_filename():
    if len(sys.argv) > 1 and sys.argv[1].endswith('.txt'):
        filename = sys.argv[1]
    else:
        print('\033[91m' + 'Provide series_matrix.txt!')
        sys.exit()
    return (filename)


def define_dividers(filename):
    study_end = 0
    samples_end = 0
    with open(filename) as file:
        for line in file:
            study_end += 1
            samples_end += 1
            if line.startswith("!Sample_title"):
                break

        for line in file:
            samples_end += 1
            if line.startswith("!series_matrix_table_begin"):
                break
    return (study_end, samples_end)


def column_merge(x): return '|'.join(x[x.notnull()].astype(str))


def create_study(filename, geo_accession, study_end):
    study = pd.read_csv(filename, sep='\t', nrows=study_end - 3)
    study["!Series_title"] = study["!Series_title"].str.replace("!", "")
    study = study.set_index("!Series_title").T
    study['Study Title'] = study.index
    study = study.groupby(level=0, axis=1).apply(lambda x: x.apply(column_merge, axis=1))
    drop_cols = []
    for i in study.columns:
        str_len = len(study[i][0])
        if str_len > 2000:
            drop_cols.append(i)
    if len(drop_cols) > 0:
        print("In study metadata following columns deleted for having too long values:", drop_cols)
    study = study.drop(drop_cols, axis=1)
    study.to_csv(geo_accession + "/" + geo_accession + "_study.tsv", sep='\t', index=False)


def create_samples(filename, geo_accession, study_end, samples_end):
    samples = pd.read_csv(filename, sep='\t', skiprows=study_end, nrows=samples_end - study_end - 2)
    samples.columns = samples.columns.str.replace('!', '')
    samples["Sample_geo_accession"] = samples["Sample_geo_accession"].str.replace("!", "")
    samples = samples.rename(columns={"Sample_geo_accession": "Sample Source ID"})
    samples = samples.set_index("Sample Source ID").T
    new_columns = []
    for item in samples.columns:
        counter = 0
        newitem = item
        while newitem in new_columns:
            counter += 1
            newitem = "{}_{}".format(item, counter)
        new_columns.append(newitem)
    samples.columns = new_columns
    samples['Sample Source ID'] = samples.index
    samples.to_csv(geo_accession + "/" + geo_accession + "_samples.tsv", sep='\t', index=False)


def create_expression(filename, geo_accession, samples_end):
    try:
        expression = pd.read_csv(filename, sep='\t', skiprows=samples_end, low_memory=False)
    except:
        return (-1)
    expression = expression.dropna()
    expression.to_csv(geo_accession + "/" + geo_accession + "_expression.tsv", sep='\t',
                      index=False)

    gct_filename = geo_accession + "_expression.gct"
    df = pd.read_csv(geo_accession + "/" + geo_accession + "_expression.tsv", sep='\t')
    df.insert(1, 'Description', None)
    df = df.rename(columns={
        df.columns[0]: "NAME"})
    rows = len(df.index)
    samples = len(df.iloc[0])

    with open(geo_accession + "/" + gct_filename, 'w') as out_file:
        writer = csv.writer(out_file, delimiter='\t')
        row = [None] * samples
        row[0] = '#1.2'
        writer.writerow(row)
        row[0], row[1] = rows, samples - 2
        writer.writerow(row)

    df.to_csv(geo_accession + "/" + gct_filename, mode='a', sep='\t', index=False)


def main():
    filename = get_filename()
    geo_accession = re.findall(r"GSE[0-9]+", filename)[0]
    print(geo_accession)
    os.makedirs(geo_accession, exist_ok=True)
    study_end, samples_end = define_dividers(filename)
    try:
        create_study(filename, geo_accession, study_end)
    except:
        print('\033[91m' + "Error creating study file!")
        sys.exit(1)
    try:
        create_samples(filename, geo_accession, study_end, samples_end)
    except:
        print('\033[91m' + "Error creating samples file!")
        sys.exit(2)
    try:
        create_expression(filename, geo_accession, samples_end)
    except:
        print("Expression not found in file. Only study and samples will be created")
    print('\033[92m' + "Files successfully prepared. Look at folder", geo_accession)

    if os.path.exists(geo_accession + "/" + geo_accession + "_expression.tsv"):
        os.remove(geo_accession + "/" + geo_accession + "_expression.tsv")


if __name__ == '__main__':
    main()
